# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional

if TYPE_CHECKING:
    from mistralai.models import (
        ChatCompletionResponse,
        Messages,
    )

from camel.configs import MISTRAL_API_PARAMS
from camel.messages import OpenAIMessage
from camel.models import BaseModelBackend
from camel.types import ChatCompletion, ModelType
from camel.utils import (
    BaseTokenCounter,
    OpenAITokenCounter,
    api_keys_required,
)

try:
    import os

    if os.getenv("AGENTOPS_API_KEY") is not None:
        from agentops import LLMEvent, record
    else:
        raise ImportError
except (ImportError, AttributeError):
    LLMEvent = None


class MistralModel(BaseModelBackend):
    r"""Mistral API in a unified BaseModelBackend interface."""

    def __init__(
        self,
        model_type: ModelType,
        model_config_dict: Dict[str, Any],
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        token_counter: Optional[BaseTokenCounter] = None,
    ) -> None:
        r"""Constructor for Mistral backend.

        Args:
            model_type (ModelType): Model for which a backend is created,
                one of MISTRAL_* series.
            model_config_dict (Dict[str, Any]): A dictionary that will
                be fed into `MistralClient.chat`.
            api_key (Optional[str]): The API key for authenticating with the
                mistral service. (default: :obj:`None`)
            url (Optional[str]): The url to the mistral service.
            token_counter (Optional[BaseTokenCounter]): Token counter to use
                for the model. If not provided, `OpenAITokenCounter` will be
                used.
        """
        super().__init__(
            model_type, model_config_dict, api_key, url, token_counter
        )
        self._api_key = api_key or os.environ.get("MISTRAL_API_KEY")
        self._url = url or os.environ.get("MISTRAL_SERVER_URL")

        from mistralai import Mistral

        self._client = Mistral(api_key=self._api_key, server_url=self._url)
        self._token_counter: Optional[BaseTokenCounter] = None

    def _to_openai_response(
        self, response: 'ChatCompletionResponse'
    ) -> ChatCompletion:
        tool_calls = None
        if (
            response.choices
            and response.choices[0].message
            and response.choices[0].message.tool_calls is not None
        ):
            tool_calls = [
                dict(
                    id=tool_call.id,  # type: ignore[union-attr]
                    function={
                        "name": tool_call.function.name,  # type: ignore[union-attr]
                        "arguments": tool_call.function.arguments,  # type: ignore[union-attr]
                    },
                    type=tool_call.TYPE,  # type: ignore[union-attr]
                )
                for tool_call in response.choices[0].message.tool_calls
            ]

        obj = ChatCompletion.construct(
            id=response.id,
            choices=[
                dict(
                    index=response.choices[0].index,  # type: ignore[index]
                    message={
                        "role": response.choices[0].message.role,  # type: ignore[index,union-attr]
                        "content": response.choices[0].message.content,  # type: ignore[index,union-attr]
                        "tool_calls": tool_calls,
                    },
                    finish_reason=response.choices[0].finish_reason  # type: ignore[index]
                    if response.choices[0].finish_reason  # type: ignore[index]
                    else None,
                )
            ],
            created=response.created,
            model=response.model,
            object="chat.completion",
            usage=response.usage,
        )

        return obj

    def _to_mistral_chatmessage(
        self,
        messages: List[OpenAIMessage],
    ) -> List["Messages"]:
        import uuid

        from mistralai.models import (
            AssistantMessage,
            FunctionCall,
            SystemMessage,
            ToolCall,
            ToolMessage,
            UserMessage,
        )

        new_messages = []
        for msg in messages:
            tool_id = uuid.uuid4().hex[:9]
            tool_call_id = uuid.uuid4().hex[:9]

            role = msg.get("role")
            function_call = msg.get("function_call")
            content = msg.get("content")

            mistral_function_call = None
            if function_call:
                mistral_function_call = FunctionCall(
                    name=function_call.get("name"),  # type: ignore[attr-defined]
                    arguments=function_call.get("arguments"),  # type: ignore[attr-defined]
                )

            tool_calls = None
            if mistral_function_call:
                tool_calls = [
                    ToolCall(function=mistral_function_call, id=tool_id)
                ]

            if role == "user":
                new_messages.append(UserMessage(content=content))  # type: ignore[arg-type]
            elif role == "assistant":
                new_messages.append(
                    AssistantMessage(content=content, tool_calls=tool_calls)  # type: ignore[arg-type]
                )
            elif role == "system":
                new_messages.append(SystemMessage(content=content))  # type: ignore[arg-type]
            elif role in {"tool", "function"}:
                new_messages.append(
                    ToolMessage(
                        content=content,  # type: ignore[arg-type]
                        tool_call_id=tool_call_id,
                        name=msg.get("name"),  # type: ignore[arg-type]
                    )
                )
            else:
                raise ValueError(f"Unsupported message role: {role}")

        return new_messages  # type: ignore[return-value]

    @property
    def token_counter(self) -> BaseTokenCounter:
        r"""Initialize the token counter for the model backend.

        # NOTE: Temporarily using `OpenAITokenCounter` due to a current issue
        # with installing `mistral-common` alongside `mistralai`.
        # Refer to: https://github.com/mistralai/mistral-common/issues/37

        Returns:
            BaseTokenCounter: The token counter following the model's
                tokenization style.
        """
        if not self._token_counter:
            self._token_counter = OpenAITokenCounter(
                model=ModelType.GPT_4O_MINI
            )
        return self._token_counter

    @api_keys_required("MISTRAL_API_KEY")
    def run(
        self,
        messages: List[OpenAIMessage],
    ) -> ChatCompletion:
        r"""Runs inference of Mistral chat completion.

        Args:
            messages (List[OpenAIMessage]): Message list with the chat history
                in OpenAI API format.

        Returns:
            ChatCompletion.
        """
        mistral_messages = self._to_mistral_chatmessage(messages)

        response = self._client.chat.complete(
            messages=mistral_messages,
            model=self.model_type.value,
            **self.model_config_dict,
        )

        openai_response = self._to_openai_response(response)  # type: ignore[arg-type]

        # Add AgentOps LLM Event tracking
        if LLMEvent:
            llm_event = LLMEvent(
                thread_id=openai_response.id,
                prompt=" ".join(
                    [message.get("content") for message in messages]  # type: ignore[misc]
                ),
                prompt_tokens=openai_response.usage.prompt_tokens,  # type: ignore[union-attr]
                completion=openai_response.choices[0].message.content,
                completion_tokens=openai_response.usage.completion_tokens,  # type: ignore[union-attr]
                model=self.model_type.value,
            )
            record(llm_event)

        return openai_response

    def check_model_config(self):
        r"""Check whether the model configuration contains any
        unexpected arguments to Mistral API.

        Raises:
            ValueError: If the model configuration dictionary contains any
                unexpected arguments to Mistral API.
        """
        for param in self.model_config_dict:
            if param not in MISTRAL_API_PARAMS:
                raise ValueError(
                    f"Unexpected argument `{param}` is "
                    "input into Mistral model backend."
                )

    @property
    def stream(self) -> bool:
        r"""Returns whether the model is in stream mode, which sends partial
        results each time. Current it's not supported.

        Returns:
            bool: Whether the model is in stream mode.
        """
        return False



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from __future__ import annotations

import asyncio
from collections import deque
from typing import Deque, Dict, List, Optional

from colorama import Fore

from camel.agents import ChatAgent
from camel.messages.base import BaseMessage
from camel.tasks.task import Task, TaskState
from camel.workforce.base import BaseNode
from camel.workforce.single_agent_node import SingleAgentNode
from camel.workforce.task_channel import TaskChannel
from camel.workforce.utils import (
    check_if_running,
    parse_assign_task_resp,
    parse_create_node_resp,
)
from camel.workforce.worker_node import WorkerNode
from camel.workforce.workforce_prompt import (
    ASSIGN_TASK_PROMPT,
    CREATE_NODE_PROMPT,
)


class ManagerNode(BaseNode):
    r"""A node that manages multiple nodes. It will split the task it
    receives into subtasks and assign them to the child nodes under
    it, and also handles the situation when the task fails.

    Args:
        description (str): Description of the node.
        coordinator_agent_kwargs (Optional[Dict]): Keyword arguments for the
            coordinator agent, e.g. `model`, `api_key`, `tools`, etc.
        task_agent_kwargs (Optional[Dict]): Keyword arguments for the task
            agent, e.g. `model`, `api_key`, `tools`, etc.
    """

    def __init__(
        self,
        description: str,
        children: List[BaseNode],
        coordinator_agent_kwargs: Optional[Dict] = None,
        task_agent_kwargs: Optional[Dict] = None,
    ) -> None:
        super().__init__(description)
        self._child_listening_tasks: Deque[asyncio.Task] = deque()
        self._children = children

        coord_agent_sysmsg = BaseMessage.make_assistant_message(
            role_name="Workforce Manager",
            content="You are coordinating a group of workers. A worker can be "
            "a group of agents or a single agent. Each worker is created to"
            " solve a specific kind of task. Your job includes assigning "
            "tasks to a existing worker, creating a new worker for a task, "
            "etc.",
        )
        self.coordinator_agent = ChatAgent(
            coord_agent_sysmsg, **(coordinator_agent_kwargs or {})
        )

        task_sys_msg = BaseMessage.make_assistant_message(
            role_name="Task Planner",
            content="You are going to compose and decompose tasks.",
        )
        self.task_agent = ChatAgent(task_sys_msg, **(task_agent_kwargs or {}))

        # if there is one, will set by the workforce class wrapping this
        self._task: Optional[Task] = None
        self._pending_tasks: Deque[Task] = deque()

    @check_if_running(False)
    def set_main_task(self, task: Task) -> None:
        r"""Set the main task for the node."""
        self._task = task

    def _get_child_nodes_info(self) -> str:
        r"""Get the information of all the child nodes under this node."""
        return '\n'.join(
            f'{child.node_id}: {child.description}' for child in self._children
        )

    def _find_assignee(
        self,
        task: Task,
        failed_log: Optional[str] = None,
    ) -> str:
        r"""Assigns a task to a child node if capable, otherwise create a
        new worker node.

        Parameters:
            task (Task): The task to be assigned.
            failed_log (Optional[str]): Optional log of a previous failed
                attempt.

        Returns:
            str: ID of the assigned node.
        """
        prompt = ASSIGN_TASK_PROMPT.format(
            content=task.content,
            child_nodes_info=self._get_child_nodes_info(),
        )
        req = BaseMessage.make_user_message(
            role_name="User",
            content=prompt,
        )
        response = self.coordinator_agent.step(req)
        try:
            print(f"{Fore.YELLOW}{response.msg.content}{Fore.RESET}")
            assignee_id = parse_assign_task_resp(response.msg.content)
        except ValueError:
            assignee_id = self._create_worker_node_for_task(task).node_id
        return assignee_id

    async def _post_task(self, task: Task, assignee_id: str) -> None:
        await self._channel.post_task(task, self.node_id, assignee_id)

    async def _post_dependency(self, dependency: Task) -> None:
        await self._channel.post_dependency(dependency, self.node_id)

    def _create_worker_node_for_task(self, task: Task) -> WorkerNode:
        r"""Creates a new worker node for a given task and add it to the
        children list of this node. This is one of the actions that
        the coordinator can take when a task has failed.

        Args:
            task (Task): The task for which the worker node is created.

        Returns:
            WorkerNode: The created worker node.
        """
        prompt = CREATE_NODE_PROMPT.format(
            content=task.content,
            child_nodes_info=self._get_child_nodes_info(),
        )
        req = BaseMessage.make_user_message(
            role_name="User",
            content=prompt,
        )
        response = self.coordinator_agent.step(req)
        new_node_conf = parse_create_node_resp(response.msg.content)

        worker_sysmsg = BaseMessage.make_assistant_message(
            role_name=new_node_conf.role,
            content=new_node_conf.system,
        )

        # TODO: add a default selection of tools for the worker
        worker = ChatAgent(worker_sysmsg)

        new_node = SingleAgentNode(
            description=new_node_conf.description,
            worker=worker,
        )
        new_node.set_channel(self._channel)

        print(
            f"{Fore.GREEN}New worker node {new_node.node_id} created."
            f"{Fore.RESET}"
        )

        self._children.append(new_node)
        self._child_listening_tasks.append(
            asyncio.create_task(new_node.start())
        )
        return new_node

    async def _get_returned_task(self) -> Task:
        r"""Get the task that's published by this node and just get returned
        from the assignee.
        """
        return await self._channel.get_returned_task_by_publisher(self.node_id)

    async def _post_ready_tasks(self) -> None:
        r"""Send all the pending tasks that have all the dependencies met to
        the channel, or directly return if there is none. For now, we will
        directly send the first task in the pending list because all the tasks
        are linearly dependent."""

        if not self._pending_tasks:
            return

        ready_task = self._pending_tasks[0]

        # if the task has failed previously, just compose and send the task
        # to the channel as a dependency
        if ready_task.state == TaskState.FAILED:
            # TODO: the composing of tasks seems not work very well
            ready_task.compose(self.task_agent)
            # remove the subtasks from the channel
            for subtask in ready_task.subtasks:
                await self._channel.remove_task(subtask.id)
            # send the task to the channel as a dependency
            await self._post_dependency(ready_task)
            self._pending_tasks.popleft()
            # try to send the next task in the pending list
            await self._post_ready_tasks()
        else:
            # directly post the task to the channel if it's a new one
            # find a node to assign the task
            assignee_id = self._find_assignee(task=ready_task)
            await self._post_task(ready_task, assignee_id)

    async def _handle_failed_task(self, task: Task) -> None:
        # remove the failed task from the channel
        await self._channel.remove_task(task.id)
        if task.get_depth() >= 3:
            # create a new WF and reassign
            # TODO: add a state for reassign?
            assignee = self._create_worker_node_for_task(task)
            # print('create_new_assignee:', assignee)
            await self._post_task(task, assignee.node_id)
        else:
            subtasks = task.decompose(self.task_agent)
            # Insert packets at the head of the queue
            self._pending_tasks.extendleft(reversed(subtasks))
            await self._post_ready_tasks()

    async def _handle_completed_task(self, task: Task) -> None:
        # archive the packet, making it into a dependency
        self._pending_tasks.popleft()
        await self._channel.archive_task(task.id)
        await self._post_ready_tasks()

    @check_if_running(False)
    def set_channel(self, channel: TaskChannel):
        r"""Set the channel for the node and all the child nodes under it."""
        self._channel = channel
        for child in self._children:
            child.set_channel(channel)

    @check_if_running(False)
    async def _listen_to_channel(self) -> None:
        r"""Continuously listen to the channel, post task to the channel and
        track the status of posted tasks.
        """

        self._running = True
        print(f"{Fore.GREEN}Manager node {self.node_id} started.{Fore.RESET}")

        # if this node is at the top level, it will have an initial task
        # the initial task must be decomposed into subtasks first
        if self._task is not None:
            subtasks = self._task.decompose(self.task_agent)
            self._pending_tasks.extend(subtasks)
            self._task.state = TaskState.FAILED
            self._pending_tasks.append(self._task)

        # before starting the loop, send ready pending tasks to the channel
        await self._post_ready_tasks()

        await self._channel.print_channel()

        while self._task is None or self._pending_tasks:
            returned_task = await self._get_returned_task()
            if returned_task.state == TaskState.DONE:
                await self._handle_completed_task(returned_task)
            elif returned_task.state == TaskState.FAILED:
                await self._handle_failed_task(returned_task)
            else:
                raise ValueError(
                    f"Task {returned_task.id} has an unexpected state."
                )

        # shut down the whole workforce tree
        self.stop()

    @check_if_running(False)
    async def start(self) -> None:
        r"""Start itself and all the child nodes under it."""
        for child in self._children:
            child_listening_task = asyncio.create_task(child.start())
            self._child_listening_tasks.append(child_listening_task)
        await self._listen_to_channel()

    @check_if_running(True)
    def stop(self) -> None:
        r"""Stop all the child nodes under it. The node itself will be stopped
        by its parent node.
        """
        for child in self._children:
            child.stop()
        for child_task in self._child_listening_tasks:
            child_task.cancel()
        self._running = False



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from __future__ import annotations

from typing import List

from camel.agents.base import BaseAgent
from camel.messages.base import BaseMessage
from camel.tasks.task import Task, TaskState
from camel.workforce.utils import parse_task_result_resp
from camel.workforce.worker_node import WorkerNode
from camel.workforce.workforce_prompt import PROCESS_TASK_PROMPT


class SingleAgentNode(WorkerNode):
    r"""A worker node that consists of a single agent.

    Args:
        description (str): Description of the node.
        worker (BaseAgent): Worker of the node. A single agent.
    """

    def __init__(
        self,
        description: str,
        worker: BaseAgent,
    ) -> None:
        super().__init__(description)
        self.worker = worker

    async def _process_task(
        self, task: Task, dependencies: List[Task]
    ) -> TaskState:
        r"""Processes a task with its dependencies.

        This method asynchronously processes a given task, considering its
        dependencies, by sending a generated prompt to a worker. It updates
        the task's result based on the agent's response.

        Args:
            task (Task): The task to process, which includes necessary details
                like content and type.
            dependencies (List[Task]): Tasks that the given task depends on.

        Returns:
            TaskState: `TaskState.DONE` if processed successfully, otherwise
                `TaskState.FAILED`.
        """
        try:
            dependency_tasks_info = self._get_dep_tasks_info(dependencies)
            prompt = PROCESS_TASK_PROMPT.format(
                content=task.content,
                type=task.type,
                dependency_task_info=dependency_tasks_info,
            )
            req = BaseMessage.make_user_message(
                role_name="User",
                content=prompt,
            )
            response = self.worker.step(req)
            # print("info['tool_calls']:", response.info['tool_calls'])
            task.result = parse_task_result_resp(response.msg.content)
            print('Task result:', task.result, '\n')
            return TaskState.DONE
        except Exception:
            return TaskState.FAILED



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
import re
from functools import wraps
from typing import Callable


class NodeConf:
    def __init__(self, role: str, system: str, description: str):
        self.role = role
        self.system = system
        self.description = description


# TODO: integrate structured response directly instead of parsing
def parse_create_node_resp(response: str) -> NodeConf:
    r"""Parses the response of the new workforce creation from the manager
    agent."""
    config = re.search(r"(<workforce>.*</workforce>)", response, re.DOTALL)
    if config is None:
        raise ValueError("No workforce configuration found in the response.")
    config_raw = config.group(1)

    try:
        import xml.etree.ElementTree as ET

        root = ET.fromstring(config_raw)
        workforce_info = {child.tag: child.text for child in root}
    except Exception as e:
        raise ValueError(f"Failed to parse workforce configuration: {e}")

    if (
        "role" not in workforce_info
        or "system" not in workforce_info
        or "description" not in workforce_info
    ):
        raise ValueError("Missing required fields in workforce configuration.")

    return NodeConf(
        role=workforce_info["role"] or "",
        system=workforce_info["system"] or "",
        description=workforce_info["description"] or "",
    )


def parse_assign_task_resp(response: str) -> str:
    r"""Parses the response of the task assignment from the manager agent."""
    assignee_id = re.search(r"<id>(.*)</id>", response)
    if assignee_id is None:
        raise ValueError("No assignee found in the response.")
    return assignee_id.group(1)


def parse_task_result_resp(response: str) -> str:
    r"""Parses the result of the task from the signle agent workforce."""
    task_result = re.search(r"<result>(.*)</result>", response, re.DOTALL)
    failed_tag = re.search(r"<failed></failed>", response)
    if failed_tag:
        task_result = None
    if task_result is None:
        raise ValueError("No result found in the response.")
    return task_result.group(1)


def check_if_running(running: bool) -> Callable:
    r"""Check if the workforce is (not) running, specified the boolean value.
    If the workforce is not in the expected status, raise an exception.

    Raises:
        RuntimeError: If the workforce is not in the expected status.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if self._running != running:
                status = "not running" if running else "running"
                raise RuntimeError(
                    f"The workforce is {status}. Cannot perform the "
                    f"operation {func.__name__}."
                )
            return func(self, *args, **kwargs)

        return wrapper

    return decorator



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from typing import List, Union

from camel.retrievers import AutoRetriever
from camel.toolkits import OpenAIFunction
from camel.toolkits.base import BaseToolkit
from camel.types import StorageType


class RetrievalToolkit(BaseToolkit):
    r"""A class representing a toolkit for information retrieval.

    This class provides methods for retrieving information from a local vector
    storage system based on a specified query.
    """

    def information_retrieval(
        self, query: str, contents: Union[str, List[str]]
    ) -> str:
        r"""Retrieves information from a local vector storage based on the
        specified query. This function connects to a local vector storage
        system and retrieves relevant information by processing the input
        query. It is essential to use this function when the answer to a
        question requires external knowledge sources.

        Args:
            query (str): The question or query for which an answer is required.
            contents (Union[str, List[str]]): Local file paths, remote URLs or
                string contents.

        Returns:
            str: The information retrieved in response to the query, aggregated
                and formatted as a string.

        Example:
            # Retrieve information about CAMEL AI.
            information_retrieval(query = "what is CAMEL AI?",
                                contents="https://www.camel-ai.org/")
        """
        auto_retriever = AutoRetriever(
            vector_storage_local_path="camel/temp_storage",
            storage_type=StorageType.QDRANT,
        )

        retrieved_info = auto_retriever.run_vector_retriever(
            query=query, contents=contents, top_k=3
        )
        return str(retrieved_info)

    def get_tools(self) -> List[OpenAIFunction]:
        r"""Returns a list of OpenAIFunction objects representing the
        functions in the toolkit.

        Returns:
            List[OpenAIFunction]: A list of OpenAIFunction objects
                representing the functions in the toolkit.
        """
        return [
            OpenAIFunction(self.information_retrieval),
        ]


# add the function to OpenAIFunction list
RETRIEVAL_FUNCS: List[OpenAIFunction] = RetrievalToolkit().get_tools()



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from abc import ABC, abstractmethod
from typing import Any

from camel.workforce.task_channel import TaskChannel


class BaseNode(ABC):
    def __init__(self, description: str) -> None:
        self.node_id = str(id(self))
        self.description = description
        # every node is initialized to use its own channel
        self._channel: TaskChannel = TaskChannel()
        self._running = False

    def reset(self, *args: Any, **kwargs: Any) -> Any:
        """Resets the node to its initial state."""
        raise NotImplementedError()

    @abstractmethod
    def set_channel(self, channel: TaskChannel):
        r"""Sets the channel for the node."""

    @abstractmethod
    async def _listen_to_channel(self):
        r"""Listens to the channel and handle tasks. This method should be
        the main loop for the node.
        """

    @abstractmethod
    async def start(self):
        r"""Start the node."""

    @abstractmethod
    def stop(self):
        r"""
        Stop the node.
        """



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from __future__ import annotations

import os
from typing import Any

from camel.embeddings.base import BaseEmbedding
from camel.types import EmbeddingModelType
from camel.utils import api_keys_required


class MistralEmbedding(BaseEmbedding[str]):
    r"""Provides text embedding functionalities using Mistral's models.

    Args:
        model_type (EmbeddingModelType, optional): The model type to be
            used for text embeddings.
            (default: :obj:`MISTRAL_EMBED`)
        api_key (str, optional): The API key for authenticating with the
            Mistral service. (default: :obj:`None`)
        dimensions (int, optional): The text embedding output dimensions.
            (default: :obj:`None`)

    Raises:
        RuntimeError: If an unsupported model type is specified.
    """

    def __init__(
        self,
        model_type: EmbeddingModelType = (EmbeddingModelType.MISTRAL_EMBED),
        api_key: str | None = None,
        dimensions: int | None = None,
    ) -> None:
        from mistralai import Mistral

        if not model_type.is_mistral:
            raise ValueError("Invalid Mistral embedding model type.")
        self.model_type = model_type
        if dimensions is None:
            self.output_dim = model_type.output_dim
        else:
            assert isinstance(dimensions, int)
            self.output_dim = dimensions
        self._api_key = api_key or os.environ.get("MISTRAL_API_KEY")
        self._client = Mistral(api_key=self._api_key)

    @api_keys_required("MISTRAL_API_KEY")
    def embed_list(
        self,
        objs: list[str],
        **kwargs: Any,
    ) -> list[list[float]]:
        r"""Generates embeddings for the given texts.

        Args:
            objs (list[str]): The texts for which to generate the embeddings.
            **kwargs (Any): Extra kwargs passed to the embedding API.

        Returns:
            list[list[float]]: A list that represents the generated embedding
                as a list of floating-point numbers.
        """
        # TODO: count tokens
        response = self._client.embeddings.create(
            inputs=objs,
            model=self.model_type.value,
            **kwargs,
        )
        return [data.embedding for data in response.data]  # type: ignore[misc,union-attr]

    def get_output_dim(self) -> int:
        r"""Returns the output dimension of the embeddings.

        Returns:
            int: The dimensionality of the embedding for the current model.
        """
        return self.output_dim



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
import asyncio

from camel.tasks import Task
from camel.workforce.manager_node import ManagerNode
from camel.workforce.task_channel import TaskChannel


class Workforce:
    r"""A class representing a workforce system.

    Args:
        name (str, optional): The name of the workforce system. Defaults to
            `"CAMEL Workforce"`.
        description (str, optional): A description of the workforce system.
            Defaults to `"A workforce system for managing tasks."`.
    """

    def __init__(
        self,
        root_node: ManagerNode,
        name: str = "CAMEL Workforce",
        description: str = "A workforce system for managing tasks.",
    ) -> None:
        self.name = name
        self.description = description
        self._root_node = root_node

    def process_task(self, task: Task) -> Task:
        self._root_node.set_main_task(task)
        shared_channel = TaskChannel()
        self._root_node.set_channel(shared_channel)

        # start the root workforce
        asyncio.run(self._root_node.start())

        return task



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from __future__ import annotations

from typing import Dict, List, Optional

from colorama import Fore, Style

from camel.agents.chat_agent import ChatAgent, FunctionCallingRecord
from camel.messages.base import BaseMessage
from camel.societies import RolePlaying
from camel.tasks.task import Task, TaskState
from camel.utils import print_text_animated
from camel.workforce.utils import parse_task_result_resp
from camel.workforce.worker_node import WorkerNode
from camel.workforce.workforce_prompt import (
    ROLEPLAY_PROCESS_TASK_PROMPT,
    ROLEPLAY_SUMMERIZE_PROMPT,
)


class RolePlayingNode(WorkerNode):
    r"""A worker node that contains a role playing.

    Args:
        description (str): Description of the node.
        assistant_role_name (str): The role name of the assistant agent.
        user_role_name (str): The role name of the user agent.
        assistant_agent_kwargs (Optional[Dict], optional): The keyword
            arguments to initialize the assistant agent in the role playing,
            like the model name, etc. Defaults to None.
        user_agent_kwargs (Optional[Dict], optional): The keyword arguments to
            initialize the user agent in the role playing, like the model name,
            etc. Defaults to None.
        chat_turn_limit (int, optional): The maximum number of chat turns in
            the role playing. Defaults to 3.
    """

    def __init__(
        self,
        description: str,
        assistant_role_name: str,
        user_role_name: str,
        assistant_agent_kwargs: Optional[Dict] = None,
        user_agent_kwargs: Optional[Dict] = None,
        chat_turn_limit: int = 3,
    ) -> None:
        super().__init__(description)
        sys_message = BaseMessage.make_assistant_message(
            role_name="Summarizer",
            content="Good at summarizing the results of the chat logs",
        )
        self.summerize_agent = ChatAgent(sys_message)
        self.chat_turn_limit = chat_turn_limit
        self.assistant_role_name = assistant_role_name
        self.user_role_name = user_role_name
        self.assistant_agent_kwargs = assistant_agent_kwargs
        self.user_agent_kwargs = user_agent_kwargs

    async def _process_task(
        self, task: Task, dependencies: List[Task]
    ) -> TaskState:
        r"""Processes a task leveraging its dependencies through role-playing.

        This method orchestrates a role-playing session between an AI
        assistant and an AI user to process a given task. It initiates with a
        generated prompt based on the task and its dependencies, conducts a
        dialogue up to a specified chat turn limit, and then summarizes the
        dialogue to determine the task's outcome.

        Args:
            task (Task): The task object to be processed, containing necessary
                details like content and type.
            dependencies (List[Task]): A list of task objects that the current
                task depends on.

        Returns:
            TaskState: `TaskState.DONE` if processed successfully, otherwise
                `TaskState.FAILED`.
        """
        try:
            dependency_tasks_info = self._get_dep_tasks_info(dependencies)
            prompt = ROLEPLAY_PROCESS_TASK_PROMPT.format(
                content=task.content,
                type=task.type,
                dependency_task_info=dependency_tasks_info,
            )
            role_play_session = RolePlaying(
                assistant_role_name=self.assistant_role_name,
                user_role_name=self.user_role_name,
                assistant_agent_kwargs=self.assistant_agent_kwargs,
                user_agent_kwargs=self.user_agent_kwargs,
                task_prompt=prompt,
                with_task_specify=False,
            )
            n = 0
            input_msg = role_play_session.init_chat()
            chat_history = []
            while n < self.chat_turn_limit:
                n += 1
                assistant_response, user_response = role_play_session.step(
                    input_msg
                )

                if assistant_response.terminated:
                    reason = assistant_response.info['termination_reasons']
                    print(
                        Fore.GREEN + f"AI Assistant terminated. Reason: "
                        f"{reason}. "
                    )
                    break
                if user_response.terminated:
                    reason = user_response.info['termination_reasons']
                    print(
                        Fore.GREEN + f"AI User terminated. Reason: {reason}."
                    )
                    break
                print_text_animated(
                    Fore.BLUE + f"AI User:\n\n{user_response.msg.content}\n",
                    delay=0.005,
                )
                chat_history.append(f"AI User: {user_response.msg.content}")

                print_text_animated(Fore.GREEN + "AI Assistant:", delay=0.005)
                tool_calls: List[FunctionCallingRecord] = (
                    assistant_response.info['tool_calls']
                )
                for func_record in tool_calls:
                    print_text_animated(f"{func_record}", delay=0.005)
                print_text_animated(
                    f"{assistant_response.msg.content}\n", delay=0.005
                )
                chat_history.append(
                    f"AI Assistant: {assistant_response.msg.content}"
                )

                if "CAMEL_TASK_DONE" in user_response.msg.content:
                    break

                input_msg = assistant_response.msg

            chat_history_str = "\n".join(chat_history)
            prompt = ROLEPLAY_SUMMERIZE_PROMPT.format(
                content=task.content,
                type=task.type,
                chat_history=chat_history_str,
            )
            req = BaseMessage.make_user_message(
                role_name="User",
                content=prompt,
            )
            response = self.summerize_agent.step(req)
            task.result = parse_task_result_resp(response.msg.content)
            print(Style.RESET_ALL + 'Task result:', task.result, '\n')
            return TaskState.DONE
        except Exception:
            return TaskState.FAILED



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========

from typing import List

from camel.utils import AgentOpsMeta

from .openai_function import OpenAIFunction


class BaseToolkit(metaclass=AgentOpsMeta):
    def get_tools(self) -> List[OpenAIFunction]:
        raise NotImplementedError("Subclasses must implement this method.")



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========

from typing import Any, Dict, List, Optional, Union

from openai import OpenAI, Stream

from camel.messages import OpenAIMessage
from camel.types import ChatCompletion, ChatCompletionChunk, ModelType
from camel.utils import (
    BaseTokenCounter,
    OpenAITokenCounter,
)


class OpenAICompatibilityModel:
    r"""Constructor for model backend supporting OpenAI compatibility."""

    def __init__(
        self,
        model_type: str,
        model_config_dict: Dict[str, Any],
        api_key: str,
        url: str,
        token_counter: Optional[BaseTokenCounter] = None,
    ) -> None:
        r"""Constructor for model backend.

        Args:
            model_type (ModelType): Model for which a backend is created.
            model_config_dict (Dict[str, Any]): A dictionary that will
                be fed into openai.ChatCompletion.create().
            api_key (str): The API key for authenticating with the
                model service. (default: :obj:`None`)
            url (str): The url to the model service. (default:
                :obj:`None`)
            token_counter (Optional[BaseTokenCounter]): Token counter to use
                for the model. If not provided, `OpenAITokenCounter(ModelType.
                GPT_3_5_TURBO)` will be used.
        """
        self.model_type = model_type
        self.model_config_dict = model_config_dict
        self._token_counter = token_counter
        self._client = OpenAI(
            timeout=60,
            max_retries=3,
            api_key=api_key,
            base_url=url,
        )

    def run(
        self,
        messages: List[OpenAIMessage],
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
        r"""Runs inference of OpenAI chat completion.

        Args:
            messages (List[OpenAIMessage]): Message list with the chat history
                in OpenAI API format.

        Returns:
            Union[ChatCompletion, Stream[ChatCompletionChunk]]:
                `ChatCompletion` in the non-stream mode, or
                `Stream[ChatCompletionChunk]` in the stream mode.
        """
        response = self._client.chat.completions.create(
            messages=messages,
            model=self.model_type,
            **self.model_config_dict,
        )
        return response

    @property
    def token_counter(self) -> BaseTokenCounter:
        r"""Initialize the token counter for the model backend.

        Returns:
            OpenAITokenCounter: The token counter following the model's
                tokenization style.
        """

        if not self._token_counter:
            self._token_counter = OpenAITokenCounter(ModelType.GPT_3_5_TURBO)
        return self._token_counter

    @property
    def stream(self) -> bool:
        r"""Returns whether the model is in stream mode, which sends partial
        results each time.

        Returns:
            bool: Whether the model is in stream mode.
        """
        return self.model_config_dict.get('stream', False)

    @property
    def token_limit(self) -> int:
        r"""Returns the maximum token limit for the given model.

        Returns:
            int: The maximum token limit for the given model.
        """
        max_tokens = self.model_config_dict.get("max_tokens")
        if isinstance(max_tokens, int):
            return max_tokens
        print(
            "Must set `max_tokens` as an integer in `model_config_dict` when"
            " setting up the model. Using 4096 as default value."
        )
        return 4096



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
from camel.prompts import TextPrompt

# ruff: noqa: E501
CREATE_NODE_PROMPT = TextPrompt(
    """You need to use the given information to create a new workforce for solving the category of tasks of the given one.
The content of the given task is:

{content}

Following is the information of the existing workforces. The format is <ID>: <description>.

{child_nodes_info}

You must return the following information:
1. The role of the agent working in the workforce, e.g. "programmer", "researcher", "product owner". This should be enclosed within the <role></role> tags.
2. The system message that will be sent to the agent, enclosed within the <system></system> tags.
3. The description of the workforce, enclosed within the <description></description> tags.

Also, all of the info should be enclosed within a <workforce> tag. For example:

<workforce>
<role>programmer</role>
<system>You are a python programmer.</system>
<description>a python programmer.</description>
</workforce>

You should ensure that the workforce created is capable of solving all the tasks in the same category as the given one, don't make it too specific.
Also, there should be no big overlap between the new workforce and the existing ones.
The information returned should be concise and clear. Each pair of tag can only appear once.
"""
)

ASSIGN_TASK_PROMPT = TextPrompt(
    """You need to assign the task to a workforce.
The content of the task is:

{content}

Following is the information of the existing workforces. The format is <ID>: <description>.

{child_nodes_info}

You must return the ID of the workforce that you think is most capable of doing the task. The ID should be enclosed within the <id></id> tags, for example:

<id>1</id>
"""
)

PROCESS_TASK_PROMPT = TextPrompt(
    """You need to process the task. It is recommended that tools be actively called when needed.
The content of the task is:

{content}

The type of the task is:

{type}

To process this task, here is some information on several dependent tasks:

{dependency_task_info}

You must return the result of the given task. The result should be enclosed within the <result></result> tags, for example:

<result>Today, you requested information about the current weather conditions. The weather today in New York City is partly cloudy with a high of 75°F (24°C) and a low of 59°F (15°C). There is a 10 percent chance of rain with winds coming from the northeast at 5 mph. Humidity levels are at 60%. It's a perfect day to spend some time outdoors, perhaps in one of the city's beautiful parks.</result>

if you are not able to process the task, you need to return <failed></failed> tags.
"""
)


ROLEPLAY_PROCESS_TASK_PROMPT = TextPrompt(
    """You need to process the task. It is recommended that tools be actively called when needed.
The content of the task is:

{content}

The type of the task is:

{type}

To process this task, here is some information on several dependent tasks:

{dependency_task_info}

You must return the result of the given task.

"""
)

# TODO: improve prompt quality
ROLEPLAY_SUMMERIZE_PROMPT = TextPrompt(
    """You need to process the task.
The content of the task is:

{content}

The type of the task is:

{type}

To process this task, here are the chat history of AI user and AI assistant:

{chat_history}

You must summerize the chat history of the return the result of the given task. The result should be enclosed within the <result></result> tags, for example:

<result>Today, you requested information about the current weather conditions. The weather today in New York City is partly cloudy with a high of 75°F (24°C) and a low of 59°F (15°C). There is a 10 percent chance of rain with winds coming from the northeast at 5 mph. Humidity levels are at 60%. It's a perfect day to spend some time outdoors, perhaps in one of the city's beautiful parks.</result>

if you are not able to process the task, you need to return <failed></failed> tags.
"""
)



# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
# Licensed under the Apache License, Version 2.0 (the “License”);
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an “AS IS” BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =========== Copyright 2023 @ CAMEL-AI.org. All Rights Reserved. ===========
import os
import warnings
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urlparse

from camel.embeddings import BaseEmbedding, OpenAIEmbedding
from camel.loaders import UnstructuredIO
from camel.retrievers.base import BaseRetriever
from camel.storages import (
    BaseVectorStorage,
    QdrantStorage,
    VectorDBQuery,
    VectorRecord,
)

try:
    from unstructured.documents.elements import Element
except ImportError:
    Element = None

DEFAULT_TOP_K_RESULTS = 1
DEFAULT_SIMILARITY_THRESHOLD = 0.75


class VectorRetriever(BaseRetriever):
    r"""An implementation of the `BaseRetriever` by using vector storage and
    embedding model.

    This class facilitates the retriever of relevant information using a
    query-based approach, backed by vector embeddings.

    Attributes:
        embedding_model (BaseEmbedding): Embedding model used to generate
            vector embeddings.
        storage (BaseVectorStorage): Vector storage to query.
        unstructured_modules (UnstructuredIO): A module for parsing files and
            URLs and chunking content based on specified parameters.
    """

    def __init__(
        self,
        embedding_model: Optional[BaseEmbedding] = None,
        storage: Optional[BaseVectorStorage] = None,
    ) -> None:
        r"""Initializes the retriever class with an optional embedding model.

        Args:
            embedding_model (Optional[BaseEmbedding]): The embedding model
                instance. Defaults to `OpenAIEmbedding` if not provided.
            storage (BaseVectorStorage): Vector storage to query.
        """
        self.embedding_model = embedding_model or OpenAIEmbedding()
        self.storage = (
            storage
            if storage is not None
            else QdrantStorage(
                vector_dim=self.embedding_model.get_output_dim()
            )
        )
        self.uio: UnstructuredIO = UnstructuredIO()

    def process(
        self,
        content: Union[str, Element],
        chunk_type: str = "chunk_by_title",
        **kwargs: Any,
    ) -> None:
        r"""Processes content from a file or URL, divides it into chunks by
        using `Unstructured IO`, and stores their embeddings in the specified
        vector storage.

        Args:
            content (Union[str, Element]): Local file path, remote URL,
                string content or Element object.
            chunk_type (str): Type of chunking going to apply. Defaults to
                "chunk_by_title".
            **kwargs (Any): Additional keyword arguments for content parsing.
        """
        if isinstance(content, Element):
            elements = [content]
        else:
            # Check if the content is URL
            parsed_url = urlparse(content)
            is_url = all([parsed_url.scheme, parsed_url.netloc])
            if is_url or os.path.exists(content):
                elements = self.uio.parse_file_or_url(content, **kwargs) or []
            else:
                elements = [self.uio.create_element_from_text(text=content)]
        if elements:
            chunks = self.uio.chunk_elements(
                chunk_type=chunk_type, elements=elements
            )
        if not elements:
            warnings.warn(
                f"No elements were extracted from the content: {content}"
            )
            return
        # Iterate to process and store embeddings, set batch of 50
        for i in range(0, len(chunks), 50):
            batch_chunks = chunks[i : i + 50]
            batch_vectors = self.embedding_model.embed_list(
                objs=[str(chunk) for chunk in batch_chunks]
            )

            records = []
            # Prepare the payload for each vector record, includes the content
            # path, chunk metadata, and chunk text
            for vector, chunk in zip(batch_vectors, batch_chunks):
                if isinstance(content, str):
                    content_path_info = {"content path": content}
                elif isinstance(content, Element):
                    content_path_info = {
                        "content path": content.metadata.file_directory
                    }
                chunk_metadata = {"metadata": chunk.metadata.to_dict()}
                chunk_text = {"text": str(chunk)}
                combined_dict = {
                    **content_path_info,
                    **chunk_metadata,
                    **chunk_text,
                }

                records.append(
                    VectorRecord(vector=vector, payload=combined_dict)
                )

            self.storage.add(records=records)

    def query(
        self,
        query: str,
        top_k: int = DEFAULT_TOP_K_RESULTS,
        similarity_threshold: float = DEFAULT_SIMILARITY_THRESHOLD,
    ) -> List[Dict[str, Any]]:
        r"""Executes a query in vector storage and compiles the retrieved
        results into a dictionary.

        Args:
            query (str): Query string for information retriever.
            similarity_threshold (float, optional): The similarity threshold
                for filtering results. Defaults to
                `DEFAULT_SIMILARITY_THRESHOLD`.
            top_k (int, optional): The number of top results to return during
                retriever. Must be a positive integer. Defaults to 1.

        Returns:
            List[Dict[str, Any]]: Concatenated list of the query results.

        Raises:
            ValueError: If 'top_k' is less than or equal to 0, if vector
                storage is empty, if payload of vector storage is None.
        """

        if top_k <= 0:
            raise ValueError("top_k must be a positive integer.")

        # Load the storage incase it's hosted remote
        self.storage.load()

        query_vector = self.embedding_model.embed(obj=query)
        db_query = VectorDBQuery(query_vector=query_vector, top_k=top_k)
        query_results = self.storage.query(query=db_query)

        if query_results[0].record.payload is None:
            raise ValueError(
                "Payload of vector storage is None, please check the "
                "collection."
            )

        # format the results
        formatted_results = []
        for result in query_results:
            if (
                result.similarity >= similarity_threshold
                and result.record.payload is not None
            ):
                result_dict = {
                    'similarity score': str(result.similarity),
                    'content path': result.record.payload.get(
                        'content path', ''
                    ),
                    'metadata': result.record.payload.get('metadata', {}),
                    'text': result.record.payload.get('text', ''),
                }
                formatted_results.append(result_dict)

        content_path = query_results[0].record.payload.get('content path', '')

        if not formatted_results:
            return [
                {
                    'text': (
                        f"No suitable information retrieved "
                        f"from {content_path} with similarity_threshold"
                        f" = {similarity_threshold}."
                    )
                }
            ]
        return formatted_results
